import { Pool } from 'pg';
import { Kysely, NoResultError, PostgresDialect, sql, } from 'kysely';
import { consoleLogger } from '../../logger';
const transferFromDB = (o) => {
    return {
        id: o.id,
        solKey: o.sol_key,
        solTx: o.sol_tx,
        sqTx: o.sq_tx,
        createdAt: o.created_at,
        payedAt: o.payed_at,
        amountDue: BigInt(o.amount_due),
        amountPayed: o.amount_payed ? BigInt(o.amount_payed) : undefined,
        operatorRewards: o.transfer_type == 'node_operator_rewards'
            ? {
                epochId: o.epoch_id,
                startTime: o.start_time,
                endTime: o.end_time,
            }
            : undefined,
        computeClientRewards: o.transfer_type == 'open_compute_rewards'
            ? {
                epochId: o.epoch_id,
                deviceId: o.device_id,
            }
            : undefined,
        upgradeRewards: o.transfer_type == 'upgrade_rewards' && !!o.epoch_id
            ? {
                epochId: o.epoch_id,
                epochStartTime: o.epoch_start_time,
                points: BigInt(o.points),
            }
            : undefined,
    };
};
const txHandle = (ps, log = consoleLogger) => {
    let { trx, commit, rollback } = ps;
    return {
        trx,
        commit,
        rollback,
        /*async releaseLock(f: { lock: Lock; newCursor?: string }): Promise<void> {
          const now = new Date(Date.now())
          const res = await trx
            .updateTable(LOCK_TABLE)
            .set({
              ...(f.newCursor ? { cursor: f.newCursor } : {}),
              released_at: now,
            })
            .where('description', '=', f.lock)
            .executeTakeFirstOrThrow()
          log.debug(`updated ${res.numUpdatedRows} lock`)
          commit()
        },*/
    };
};
export const TRANSFER_TABLE = 'transfer';
export const NODE_OPERATOR_REWARDS_TRANSFER_TABLE = 'node_operator_rewards_transfer';
export const OPEN_COMPUTE_REWARDS_TRANSFER_TABLE = 'open_compute_rewards_transfer';
export const UPGRADE_REWARDS_TRANSFER_TABLE = 'upgrade_rewards_transfer';
export const SQUADS_TX_TABLE = 'squads_tx';
export const NODE_OPERATOR_EPOCH_TABLE = 'node_operator_epoch';
export const JOB_ID_TABLE = 'job_id';
export const JOB_TABLE = 'job';
export const BURN_TABLE = 'burn';
export const pgClient = (config) => {
    const dialect = new PostgresDialect({
        pool: new Pool({
            ...config,
            port: config.port ?? 5432,
            max: config.max ?? 10,
        }),
    });
    // Database interface is passed to Kysely's constructor, and from now on, Kysely
    // knows your database structure.
    // Dialect is passed to Kysely's constructor, and from now on, Kysely knows how
    // to communicate with your database.
    const db = new Kysely({
        dialect,
    });
    const createTransfers = async (db, transfers, log = consoleLogger) => {
        for (const t of transfers) {
            const { id: transferId } = (await db
                .insertInto(TRANSFER_TABLE)
                .values({
                amount_due: BigInt(t.amountDue),
                amount_payed: t.amountPayed ? BigInt(t.amountPayed) : 0,
                created_at: t.createdAt,
                payed_at: t.payedAt,
                sol_key: t.solKey,
                sol_tx: t.solTx,
                transfer_type: (() => {
                    if (t.operatorRewards) {
                        return 'node_operator_rewards';
                    }
                    else if (t.upgradeRewards) {
                        return 'upgrade_rewards';
                    }
                    else if (t.computeClientRewards) {
                        return 'open_compute_rewards';
                    }
                    throw new Error('wtf');
                })(),
            })
                .returning((eb) => ['id as id'])
                .executeTakeFirst());
            log.info(`inserted transfer ${JSON.stringify(t, (key, value) => typeof value === 'bigint' ? value.toString() : value)}`);
            if (t.operatorRewards) {
                log.info(`inserting operator rewards ${transferId}`);
                const x = (await db
                    .insertInto(NODE_OPERATOR_REWARDS_TRANSFER_TABLE)
                    .values({
                    end_time: t.operatorRewards.endTime,
                    start_time: t.operatorRewards.startTime,
                    epoch_id: t.operatorRewards.epochId,
                    transfer_id: transferId,
                })
                    .executeTakeFirst()).insertId;
            }
            else if (t.computeClientRewards) {
                log.info(`inserting compute client rewards ${transferId}`);
                const x = (await db
                    .insertInto(OPEN_COMPUTE_REWARDS_TRANSFER_TABLE)
                    .values({
                    epoch_id: t.computeClientRewards.epochId,
                    device_id: t.computeClientRewards.deviceId,
                    transfer_id: transferId,
                })
                    .executeTakeFirst()).insertId;
            }
            else if (t.upgradeRewards) {
                log.info(`inserting upgrade rewards ${transferId}`);
                const x = (await db
                    .insertInto(UPGRADE_REWARDS_TRANSFER_TABLE)
                    .values({
                    epoch_id: t.upgradeRewards.epochId,
                    epoch_start_time: t.upgradeRewards.epochStartTime,
                    points: t.upgradeRewards.points,
                    transfer_id: transferId,
                })
                    .executeTakeFirst()).insertId;
            }
        }
        return;
    };
    return {
        db,
        async saveNodeOperatorEpoch(epochId, transfers, jobIds, totalRndr, log = consoleLogger) {
            await db.transaction().execute(async (trx) => {
                await createTransfers(trx, transfers, log);
                log.info(`inserted transfers`);
                for (const id of jobIds) {
                    await trx
                        .insertInto(JOB_ID_TABLE)
                        .values({
                        id: BigInt(id),
                        epoch_id: epochId,
                    })
                        .executeTakeFirstOrThrow();
                }
                log.info(`inserted jobids`);
                const now = new Date(Date.now());
                const res = await trx
                    .updateTable(NODE_OPERATOR_EPOCH_TABLE)
                    .set({
                    saved_at: now,
                    total_rndr: totalRndr,
                })
                    .where('id', '=', epochId.toString())
                    .executeTakeFirstOrThrow();
                log.info(`saved node operator epoch: ${res.numUpdatedRows}`);
            });
            return;
        },
        async getLatestNodeOperatorEpoch(log = consoleLogger) {
            const now = new Date(Date.now());
            try {
                const res = await db
                    .selectFrom(NODE_OPERATOR_EPOCH_TABLE)
                    .selectAll()
                    .where('end_time', '<=', now)
                    .where('saved_at', 'is', null)
                    .orderBy('end_time asc')
                    .limit(1)
                    .executeTakeFirstOrThrow();
                return {
                    epochId: BigInt(res.id),
                    startedAt: res.start_time,
                    endedAt: res.end_time,
                    savedAt: res.saved_at ?? undefined,
                };
            }
            catch (e) {
                if (e instanceof NoResultError) {
                    log.info(`no node operator epoch found`);
                    return undefined;
                }
                throw e;
            }
        },
        async getOutstanding(f, log = consoleLogger) {
            const { typ, nullKey, due } = f;
            //select sol_key, sum(amount_due)-sum(amount_payed) from transfer group by sol_key;
            let q = db
                .selectFrom(TRANSFER_TABLE)
                .select(({ fn, val, ref }) => [
                'sol_key',
                sql `${fn.sum('amount_due')} -
          ${fn.sum('amount_payed')}`.as('due'),
            ])
                .where('transfer_type', '=', typ)
                .groupBy(['sol_key']);
            if (!nullKey) {
                q = q.where('sol_key', 'is not', null);
            }
            if (due) {
                q = q.having(({ fn, ref }) => sql `${fn.sum('amount_due')} -
            ${fn.sum('amount_payed')}
            >
            0`);
            }
            const res = await q.execute();
            log.info(`outstanding query ${q.compile().sql} params ${JSON.stringify(f)} ${res.length} rows`);
            const out = {};
            for (const r of res) {
                out[r.sol_key ?? ''] = BigInt(r.due);
            }
            return out;
        },
        async fetchTransfers(f, log = consoleLogger) {
            let base = db
                .selectFrom(TRANSFER_TABLE)
                // @ts-ignore
                .selectAll(TRANSFER_TABLE)
                .where('transfer.transfer_type', '=', f.typ);
            if (f.solKey) {
                base = base.where('transfer.sol_key', '=', f.solKey);
            }
            if (f.payed != undefined) {
                base = base.where('transfer.payed_at', f.payed ? 'is not' : 'is', null);
            }
            if (f.createdBefore != undefined) {
                base = base.where('transfer.created_at', '<=', f.createdBefore);
            }
            if (f.typ == 'upgrade_rewards') {
                base = base
                    .innerJoin(UPGRADE_REWARDS_TRANSFER_TABLE, 'upgrade_rewards_transfer.transfer_id', 'transfer.id')
                    .selectAll(UPGRADE_REWARDS_TRANSFER_TABLE);
                if (f.epochId) {
                    // @ts-ignore
                    base = base.where('upgrade_rewards_transfer.epoch_id', '=', f.epochId);
                }
            }
            if (f.typ == 'open_compute_rewards') {
                base = base
                    .innerJoin(OPEN_COMPUTE_REWARDS_TRANSFER_TABLE, 'open_compute_rewards_transfer.transfer_id', 'transfer.id')
                    .selectAll(OPEN_COMPUTE_REWARDS_TRANSFER_TABLE);
                if (f.epochId) {
                    base = base.where(
                    // @ts-ignore
                    'open_compute_rewards_transfer.epoch_id', '=', f.epochId);
                }
            }
            if (f.typ == 'node_operator_rewards') {
                base = base
                    .innerJoin(NODE_OPERATOR_REWARDS_TRANSFER_TABLE, 'node_operator_rewards_transfer.transfer_id', 'transfer.id')
                    .selectAll(NODE_OPERATOR_REWARDS_TRANSFER_TABLE);
                if (f.epochId) {
                    base = base.where(
                    // @ts-ignore
                    'node_operator_rewards_transfer.epoch_id', '=', f.epochId);
                }
            }
            if (f.sort) {
                base = base.orderBy('transfer.sol_key asc');
            }
            const res = await base.execute();
            log.info(`fetched ${base.compile().sql} params ${JSON.stringify(f)} ${res.length} transfers`);
            return res.map((r) => transferFromDB(r));
        },
        async payoutTransfers(idToAmt, payedAt, solTx, log = consoleLogger) {
            const res = await db.transaction().execute(async (tx) => {
                const alreadyUpdatedIds = (await tx
                    .selectFrom(TRANSFER_TABLE)
                    .select(['id'])
                    .where('id', 'in', Object.keys(idToAmt))
                    .where('sol_tx', 'is not', null)
                    .execute()).map(({ id }) => id);
                for (const [id, amt] of Object.entries(idToAmt)) {
                    if (alreadyUpdatedIds.includes(id)) {
                        log.info(`transfer ${id} already marked payed out. skipping`);
                        continue;
                    }
                    const res = await tx
                        .updateTable(TRANSFER_TABLE)
                        .set({
                        payed_at: payedAt,
                        sol_tx: solTx,
                        amount_payed: amt,
                    })
                        .where('id', '=', id.toString())
                        .executeTakeFirstOrThrow();
                    log.info(`marked ${res.numUpdatedRows} transfer payed out`);
                }
            });
        },
        async payoutTransfer(id, payedAt, solTx, amtPayed, log = consoleLogger) {
            const res = await db
                .updateTable(TRANSFER_TABLE)
                .set({
                payed_at: payedAt,
                sol_tx: solTx,
                amount_payed: amtPayed,
            })
                .where('id', '=', id.toString())
                .executeTakeFirst();
            log.info(`marked ${res.numUpdatedRows} transfer payed out`);
            return;
        },
        async createTransfers(transfers, log = consoleLogger) {
            await createTransfers(db, transfers, log);
        },
        async fetchInfoForEpoch(epochId) {
            const res = await db
                .selectFrom(JOB_ID_TABLE)
                .innerJoin(NODE_OPERATOR_EPOCH_TABLE, 'node_operator_epoch.id', 'job_id.epoch_id')
                .where('job_id.epoch_id', '=', epochId.toString())
                .select([
                'job_id.id',
                'node_operator_epoch.total_rndr',
                'node_operator_epoch.start_time',
                'node_operator_epoch.end_time',
                'node_operator_epoch.burned',
                'node_operator_epoch.burn_tx',
            ])
                .execute();
            if (res.length == 0) {
                throw new Error(`epoch ${epochId} not found`);
            }
            return {
                jobIds: res.map((r) => Number(r.id)),
                totalRndr: BigInt(res[0].total_rndr),
                startTime: res[0].start_time,
                endTime: res[0].end_time,
                burned: BigInt(res[0].burned),
                burnTx: res[0].burn_tx,
            };
        },
        async updateComputeClientSolKey(deviceId, solKey, log = consoleLogger) {
            const q = db
                .updateTable(`${TRANSFER_TABLE} as t`)
                .set({
                sol_key: solKey,
            })
                .from(`${OPEN_COMPUTE_REWARDS_TRANSFER_TABLE} as o`)
                .whereRef('t.id', '=', 'o.transfer_id')
                .where('t.sol_key', 'is', null)
                .where('o.device_id', '=', deviceId);
            const sql = await q.compile();
            log.info(`sql ${sql.sql} params ${sql.parameters}`);
            const res = await q.executeTakeFirstOrThrow();
            log.info(`updated ${res.numUpdatedRows} device-id sol_keys`);
        },
        async markJobsBurned(p, log = consoleLogger) {
            let { markedAt, solTx } = p;
            let res = await db
                .updateTable(BURN_TABLE)
                .set({ marked_burned_at: markedAt })
                .where('sol_tx', '=', solTx)
                .executeTakeFirstOrThrow();
            log.debug(`set ${res.numUpdatedRows} ${solTx} -> marked_burned_at ${markedAt.toISOString()}`);
        },
        async markJobsBurnt(trx, p, log = consoleLogger) {
            const { id } = await trx
                .insertInto(BURN_TABLE)
                .values({
                sol_tx: p.solTx,
                executed_at: p.executedAt,
                priced_at: p.pricedAt,
                usdc_spent: p.usdcSpent,
                quote_amt: p.quoteAmt,
                burned: p.burned,
            })
                .onConflict((oc) => oc.column('sol_tx').doNothing())
                .returning((eb) => ['id as id'])
                .executeTakeFirstOrThrow();
            const jobIdsStr = p.jobsIds.map(String);
            const res = await trx
                .updateTable(JOB_TABLE)
                .where('id', 'in', jobIdsStr)
                .set('burn_id', id)
                .executeTakeFirstOrThrow();
            log.debug(`set burn_id on ${res.numUpdatedRows} jobs`);
            return;
        },
        async produceBurnedJobs(jobs, log = consoleLogger) {
            await db.transaction().execute(async (trx) => {
                let num = BigInt(0);
                for (const { amount, completedAt, jobId } of jobs) {
                    const res = await trx
                        .insertInto(JOB_TABLE)
                        .values({
                        id: jobId,
                        render_amt: amount,
                        completed_at: completedAt,
                        inserted_at: new Date(Date.now()),
                    })
                        .onConflict((oc) => oc.column('id').doNothing())
                        .executeTakeFirstOrThrow();
                    num += res.numInsertedOrUpdatedRows ?? BigInt(0);
                }
                log.debug(`inserted ${num} of ${jobs.length} burn jobs`);
            });
        },
        async consumeAndLockBurnJobs(limit, lowerRndrThreshold, upperRndrThreshold, log = consoleLogger) {
            const trx = await begin(db);
            const q = trx.trx
                .selectFrom(trx.trx
                .selectFrom(JOB_TABLE)
                .select(({ fn, val, ref }) => [
                'id',
                fn
                    .sum('render_amt')
                    .over((ob) => ob.orderBy(`completed_at`, `asc`))
                    .as('total'),
            ])
                .where('burn_id', 'is', null)
                .as('t'))
                .select('id')
                .where('total', '>', lowerRndrThreshold)
                .where('total', '<', upperRndrThreshold);
            log.debug(`outstanding jobs q: ${q.compile().sql} ${JSON.stringify({
                limit,
                lowerRndrThreshold,
                upperRndrThreshold,
            })}`);
            const idRs = await q.execute();
            const ids = idRs.map((x) => x.id);
            log.debug(`${ids.length} candidate jobs for burn. locking up to ${limit} jobs`);
            if (ids.length == 0) {
                return { trx, jobs: [] };
            }
            const rows = await trx.trx
                .selectFrom(JOB_TABLE)
                .selectAll()
                .where('id', 'in', ids)
                .limit(limit)
                .forUpdate()
                .skipLocked()
                .execute();
            log.debug(`locked ${rows.length} jobs`);
            const jobs = [];
            for (const r of rows) {
                jobs.push({
                    id: Number(r.id),
                    completedAt: r.completed_at,
                    rndrUsed: BigInt(r.render_amt),
                });
            }
            return { trx, jobs };
        },
        async incrBuyBurnMetrics(log = consoleLogger) {
            const res = await db
                .selectFrom(JOB_TABLE)
                .select('completed_at')
                .orderBy('completed_at asc')
                .where('burn_id', 'is', null)
                .execute();
            if (res.length == 0) {
                return { outstandingJobs: 0 };
            }
            return {
                outstandingJobs: res.length,
                oldestJobFinishedAt: res[0].completed_at,
            };
        },
        async deleteAll() { },
    };
};
async function begin(db) {
    const connection = new Deferred();
    const result = new Deferred();
    // Do NOT await this line.
    db.transaction()
        .execute((trx) => {
        connection.resolve(trx);
        return result.promise;
    })
        .catch((err) => {
        // Don't do anything here. Just swallow the exception.
    });
    const trx = await connection.promise;
    return {
        trx,
        commit() {
            result.resolve(null);
        },
        rollback() {
            result.reject(new Error('rollback'));
        },
    };
}
export class Deferred {
    #promise;
    #resolve;
    #reject;
    constructor() {
        this.#promise = new Promise((resolve, reject) => {
            this.#reject = reject;
            this.#resolve = resolve;
        });
    }
    get promise() {
        return this.#promise;
    }
    resolve = (value) => {
        if (this.#resolve) {
            this.#resolve(value);
        }
    };
    reject = (reason) => {
        if (this.#reject) {
            this.#reject(reason);
        }
    };
}
//# sourceMappingURL=v1.js.map