import { DEFAULT_MULTISIG_PROGRAM_ID, getAuthorityPDA, getTxPDA, } from '@sqds/sdk';
import BN from 'bn.js';
import { ComputeBudgetProgram, Transaction, TransactionMessage, VersionedTransaction, } from '@solana/web3.js';
import { sendAndConfirmWithRetry, sendAndConfirmWithRetryAndPrioFees, } from '@render-foundation/spl-utils';
import { consoleLogger } from '../logger';
import { generated, PROGRAM_ID } from '@render-foundation/multisig';
import * as multisig from '@render-foundation/multisig';
export const executeSqV4Transaction = async (connection, multisigPda, approvers, instrs, log = consoleLogger, prioFee = 0) => {
    let computePriceIx;
    if (prioFee > 0) {
        log.debug(`adding prioFee ${prioFee}`);
        computePriceIx = ComputeBudgetProgram.setComputeUnitPrice({
            // @ts-ignore
            microLamports: Math.trunc(prioFee),
        });
    }
    let multisigAccount = await generated.Multisig.fromAccountAddress(connection, multisigPda);
    let vaultPda = multisig.getVaultPda({
        multisigPda,
        index: 0,
        programId: PROGRAM_ID,
    })[0];
    const transactionIndex = multisig.utils.toBigInt(multisigAccount.transactionIndex) + BigInt(1);
    const txPDA = multisig.getTransactionPda({
        multisigPda,
        index: transactionIndex,
    })[0];
    const txMessage = new TransactionMessage({
        payerKey: vaultPda,
        recentBlockhash: (await connection.getLatestBlockhash()).blockhash,
        instructions: instrs,
    });
    const createIx = multisig.instructions.vaultTransactionCreate({
        multisigPda,
        transactionIndex,
        creator: approvers[0].publicKey,
        vaultIndex: 0,
        ephemeralSigners: 0,
        transactionMessage: txMessage,
        memo: `execute ${txPDA}`,
        programId: PROGRAM_ID,
    });
    // Create Proposal account.
    const createPropIx = multisig.instructions.proposalCreate({
        multisigPda,
        transactionIndex,
        creator: approvers[0].publicKey,
        programId: multisig.PROGRAM_ID,
    });
    //await connection.confirmTransaction(signature);
    const createTx = new Transaction();
    if (computePriceIx) {
        createTx.add(computePriceIx);
    }
    createTx.add(createIx, createPropIx);
    for (const approver of approvers) {
        createTx.add(multisig.instructions.proposalApprove({
            multisigPda,
            transactionIndex,
            member: approver.publicKey,
            memo: `${approver.publicKey} approve`,
            programId: multisig.PROGRAM_ID,
        }));
    }
    createTx.recentBlockhash = (await connection.getLatestBlockhash()).blockhash;
    createTx.feePayer = approvers[0].publicKey;
    createTx.sign(...approvers);
    await sendAndConfirmWithRetry(connection, createTx.serialize(), {
        skipPreflight: true,
        maxRetries: 10,
    }, 'finalized', prioFee == 0 ? 120000 : undefined);
    log.info(`created tx ${txPDA}`);
    log.info(`executing tx ${txPDA}`);
    const execIx = await multisig.instructions.vaultTransactionExecute({
        connection: connection,
        multisigPda,
        transactionIndex,
        member: approvers[0].publicKey,
        programId: PROGRAM_ID,
    });
    let execIxs = [];
    if (prioFee > 0) {
        execIxs.push(computePriceIx);
    }
    execIxs.push(execIx.instruction);
    const execTx = new VersionedTransaction(new TransactionMessage({
        payerKey: vaultPda,
        recentBlockhash: (await connection.getLatestBlockhash()).blockhash,
        instructions: execIxs,
    }).compileToV0Message(execIx.lookupTableAccounts));
    /*const execTx = await multisig.transactions.vaultTransactionExecute({
      connection: connection,
      feePayer: approvers[0].publicKey,
      blockhash: (await connection.getLatestBlockhash()).blockhash,
      multisigPda,
      transactionIndex,
      member: approvers[0].publicKey,
      programId: PROGRAM_ID,
    })*/
    execTx.sign([approvers[0]]);
    const { txid } = await sendAndConfirmWithRetry(connection, Buffer.from(execTx.serialize()), {
        skipPreflight: true,
        maxRetries: 10,
    }, 'confirmed', prioFee == 0 ? 120000 : undefined);
    log.info(`executed tx ${txPDA}`);
    return { txSig: txid, txPDA };
};
export const executeSqV3Transaction = async (connection, squadsMsKey, clients, instrs, log = consoleLogger, startPrio = 'NONE') => {
    const msAuthorityK = getAuthorityPDA(squadsMsKey, new BN(1), DEFAULT_MULTISIG_PROGRAM_ID)[0];
    let txPDA;
    const primClient = clients[0];
    const txIdx = await primClient.getNextTransactionIndex(squadsMsKey);
    txPDA = getTxPDA(squadsMsKey, new BN(txIdx), DEFAULT_MULTISIG_PROGRAM_ID)[0];
    let ixs = [
        await primClient.buildCreateTransaction(squadsMsKey, 1, txIdx),
    ];
    let instrIdx = 1;
    for (const i of instrs) {
        ixs.push(await primClient.buildAddInstruction(squadsMsKey, txPDA, i, instrIdx));
        instrIdx += 1;
    }
    ixs.push(await primClient.buildActivateTransaction(squadsMsKey, txPDA));
    for (const c of clients) {
        ixs.push(await c.buildApproveTransaction(squadsMsKey, txPDA));
    }
    //createTx.recentBlockhash = (await connection.getLatestBlockhash()).blockhash
    //createTx.feePayer = primClient.wallet.publicKey
    /*for (const c of clients) {
      createTx = await c.wallet.signTransaction(createTx)
    }*/
    //createTx.sign(sharedMsKp, authorityKp)
    await sendAndConfirmWithRetryAndPrioFees(connection, ixs, clients.map((c) => c.wallet.payer), {
        skipPreflight: true,
        maxRetries: 10,
    }, 'confirmed', primClient.wallet.payer.publicKey, startPrio, undefined, log);
    log.info(`created tx ${txPDA}`);
    log.info(`executing tx ${txPDA}`);
    //let execTx = new Transaction().add()
    //execTx.recentBlockhash = (await connection.getLatestBlockhash()).blockhash
    //execTx.feePayer = primClient.wallet.publicKey
    //execTx = await primClient.wallet.signTransaction(execTx)
    //execTx.sign(args.solFeePayerKp)
    const { txid } = await sendAndConfirmWithRetryAndPrioFees(connection, [
        await primClient.buildExecuteTransaction(txPDA, primClient.wallet.publicKey),
    ], [primClient.wallet.payer], {
        skipPreflight: true,
        maxRetries: 10,
    }, 'confirmed', primClient.wallet.payer.publicKey, startPrio, undefined, log);
    log.info(`executed ${txid}`);
    return { txSig: txid, txPDA };
};
//# sourceMappingURL=squads.js.map