import { DB } from '../../dbTypes';
import { Kysely, Transaction } from 'kysely';
import { RenderLogger } from '../../logger';
export type { DB } from '../../dbTypes';
export type TransferType = 'open_compute_rewards' | 'node_operator_rewards' | 'upgrade_rewards';
export interface Transfer {
    id?: bigint;
    solKey?: string;
    solTx?: string;
    sqTx?: string;
    createdAt?: Date;
    payedAt?: Date;
    amountDue: bigint;
    amountPayed?: bigint;
    operatorRewards?: NodeOperatorRewardsTransfer;
    computeClientRewards?: ComputeClientsRewardsTransfer;
    upgradeRewards?: UpgradeRewardsTransfer;
}
export interface Job {
    id: number;
    completedAt: Date;
    rndrUsed: bigint;
}
export interface UpgradeRewardsTransfer {
    epochId: number;
    epochStartTime?: Date;
    points: bigint;
}
export interface NodeOperatorRewardsTransfer {
    epochId: number;
    startTime: Date;
    endTime: Date;
}
export interface ComputeClientsRewardsTransfer {
    epochId: number;
    deviceId: string;
}
export interface TransactionHandle<DB> {
    trx: Transaction<DB>;
    commit: () => void;
    rollback: () => void;
}
export interface PgClient<DB> {
    db: Kysely<DB>;
    updateComputeClientSolKey(deviceId: string, solKey: string, log?: RenderLogger): Promise<void>;
    consumeAndLockBurnJobs(limit: number, lowerRndrThreshold: number, upperRndrThreshold: number, log?: RenderLogger): Promise<{
        trx: TransactionHandle<DB>;
        jobs: Job[];
    }>;
    fetchInfoForEpoch(epochId: bigint, log?: RenderLogger): Promise<{
        jobIds: number[];
        totalRndr: bigint;
        startTime?: Date;
        endTime?: Date;
        burned: bigint;
        burnTx?: string;
    }>;
    getOutstanding(f: {
        typ: TransferType;
        nullKey?: boolean;
        due?: boolean;
    }, log?: RenderLogger): Promise<Record<string, bigint>>;
    createTransfers(transfers: Transfer[], log?: RenderLogger): Promise<void>;
    fetchTransfers(f: {
        epochId?: number;
        payed?: boolean;
        solKey?: string;
        typ: TransferType;
        createdBefore?: Date;
        sort?: 'solKey';
    }, log?: RenderLogger): Promise<Transfer[]>;
    payoutTransfer(id: bigint, payedAt: Date, solTx: string, amtPayed: bigint, log?: RenderLogger): Promise<void>;
    payoutTransfers(idToAmt: Record<string, bigint>, payedAt: Date, solTx: string, log?: RenderLogger): Promise<void>;
    getLatestNodeOperatorEpoch(log?: RenderLogger): Promise<{
        epochId: bigint;
        startedAt: Date;
        endedAt: Date;
        savedAt?: Date;
    } | undefined>;
    saveNodeOperatorEpoch(epochId: bigint, transfers: Transfer[], jobIds: number[], totalRndr: bigint, log?: RenderLogger): Promise<void>;
    produceBurnedJobs(jobs: {
        completedAt: Date;
        jobId: number;
        amount: number;
    }[], log?: RenderLogger): Promise<void>;
    markJobsBurnt(trx: Transaction<DB>, p: {
        jobsIds: number[];
        solTx: string;
        burned: number;
        usdcSpent?: number;
        quoteAmt?: number;
        pricedAt: Date;
        executedAt: Date;
    }, log?: RenderLogger): Promise<void>;
    markJobsBurned(p: {
        solTx: string;
        markedAt: Date;
    }, log?: RenderLogger): Promise<void>;
    incrBuyBurnMetrics(log?: RenderLogger): Promise<{
        outstandingJobs: number;
        oldestJobFinishedAt?: Date;
    }>;
    deleteAll(): Promise<void>;
}
export declare const TRANSFER_TABLE = "transfer";
export declare const NODE_OPERATOR_REWARDS_TRANSFER_TABLE = "node_operator_rewards_transfer";
export declare const OPEN_COMPUTE_REWARDS_TRANSFER_TABLE = "open_compute_rewards_transfer";
export declare const UPGRADE_REWARDS_TRANSFER_TABLE = "upgrade_rewards_transfer";
export declare const SQUADS_TX_TABLE = "squads_tx";
export declare const NODE_OPERATOR_EPOCH_TABLE = "node_operator_epoch";
export declare const JOB_ID_TABLE = "job_id";
export declare const JOB_TABLE = "job";
export declare const BURN_TABLE = "burn";
export interface PgConfig {
    database: string;
    host: string;
    user: string;
    password: string;
    port?: number;
    max?: number;
}
export declare const pgClient: (config: PgConfig) => PgClient<DB>;
export declare class Deferred<T> {
    #private;
    constructor();
    get promise(): Promise<T>;
    resolve: (value: T | PromiseLike<T>) => void;
    reject: (reason?: any) => void;
}
//# sourceMappingURL=v1.d.ts.map